@@ -1,16 +1,16 @@
 use strict;
 use warnings;
-package JSON::Schema::Draft201909; # git description: v0.001-20-gbe3f104
+package JSON::Schema::Draft201909; # git description: v0.002-24-g55b80c8
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate data against a schema
 # KEYWORDS: JSON Schema data validation structure specification
 
-our $VERSION = '0.002';
+our $VERSION = '0.003';
 
 no if "$]" >= 5.031009, feature => 'indirect';
 use feature 'current_sub';
 use JSON::MaybeXS 1.004001 'is_bool';
-use Syntax::Keyword::Try;
+use Syntax::Keyword::Try 0.11;
 use Carp 'croak';
 use List::Util 1.33 qw(any pairs);
 use Mojo::JSON::Pointer;
@@ -19,7 +19,7 @@ use Safe::Isa;
 use Moo;
 use MooX::TypeTiny 0.002002;
 use MooX::HandlesVia;
-use Types::Standard 1.010002 qw(Bool HasMethods Enum InstanceOf HashRef Dict);
+use Types::Standard 1.010002 qw(Bool Int HasMethods Enum InstanceOf HashRef Dict);
 use JSON::Schema::Draft201909::Error;
 use JSON::Schema::Draft201909::Result;
 use namespace::clean;
@@ -37,6 +37,12 @@ has short_circuit => (
   default => sub { $_[0]->output_format eq 'flag' },
 );
 
+has max_traversal_depth => (
+  is => 'ro',
+  isa => Int,
+  default => 50,
+);
+
 has _resource_index => (
   is => 'bare',
   isa => HashRef[Dict[
@@ -68,8 +74,12 @@ before _add_resources => sub {
             or $existing->{canonical_uri} ne $value->{canonical_uri};
     }
 
+    my $fragment = $value->{canonical_uri}->fragment;
+    croak sprintf('canonical_uri cannot contain an empty fragment (%s)', $value->{canonical_uri})
+      if defined $fragment and $fragment eq '';
+
     croak sprintf('canonical_uri cannot contain a plain-name fragment (%s)', $value->{canonical_uri})
-      if ($value->{canonical_uri}->fragment // '') =~ m{^[^/]};
+      if ($fragment // '') =~ m{^[^/]};
   }
 };
 
@@ -111,9 +121,10 @@ sub evaluate {
   my $state = {
     base_uri => Mojo::URL->new,                   # TODO: will be set by a global attribute
     short_circuit => $self->short_circuit,
+    depth => 0,
     data_path => '',
     traversed_schema_path => '',  # the accumulated path up to the last $ref traversal
-    absolute_schema_uri => undef, # the absolute path of the last traversed $ref; always a Mojo::URL
+    canonical_schema_uri => undef,# the canonical path of the last traversed $ref (Mojo::URL)
     schema_path => '',            # the rest of the path, since the last traversed $ref
     errors => [],
   };
@@ -146,6 +157,9 @@ sub _eval {
   $state = { %$state };     # changes to $state should only affect subschemas, not parents
   delete $state->{keyword};
 
+  abort($state, 'maximum traversal depth exceeded')
+    if $state->{depth}++ > $self->max_traversal_depth;
+
   my $schema_type = $self->_get_type($schema);
   return $schema || E($state, 'subschema is false') if $schema_type eq 'boolean';
 
@@ -180,22 +194,10 @@ sub _eval {
   return $result;
 }
 
-sub _eval_keyword_comment {
-  my ($self, $data, $schema, $state) = @_;
-  abort($state, '"$comment" value is not a string')
-    if not $self->_is_type('string', $schema->{'$comment'});
-  # we do nothing with this keyword, including not collecting its value for annotations.
-  return 1;
-}
-
-sub _eval_keyword_defs {
-  # we do nothing directly with this keyword, including not collecting its value for annotations.
-  return 1;
-}
-
 sub _eval_keyword_schema {
   my ($self, $data, $schema, $state) = @_;
 
+  assert_keyword_type($state, $schema, 'string');
   abort($state, 'custom $schema references are not yet supported')
     if $schema->{'$schema'} ne 'https://json-schema.org/draft/2019-09/schema';
 
@@ -205,15 +207,16 @@ sub _eval_keyword_schema {
 sub _eval_keyword_id {
   my ($self, $data, $schema, $state) = @_;
 
-  abort($state, '%s is not a string', $schema->{'$id'})
-    if not $self->_is_type('string', $schema->{'$id'});
+  assert_keyword_type($state, $schema, 'string');
 
   my $uri = Mojo::URL->new($schema->{'$id'})->base($state->{base_uri})->to_abs;
-  abort($state, '%s cannot have a non-empty fragment', $schema->{'$id'}) if length $uri->fragment;
+  abort($state, '$id value "%s" cannot have a non-empty fragment', $schema->{'$id'})
+    if length $uri->fragment;
 
+  $uri->fragment(undef);
   $state->{base_uri} = $uri;
   $state->{traversed_schema_path} = $state->{traversed_schema_path}.$state->{schema_path};
-  $state->{absolute_schema_uri} = $uri->clone;
+  $state->{canonical_schema_uri} = $uri->clone;
   $state->{schema_path} = '';
 
   return 1;
@@ -222,37 +225,38 @@ sub _eval_keyword_id {
 sub _eval_keyword_anchor {
   my ($self, $data, $schema, $state) = @_;
 
-  abort($state, '%s is not a string', $schema->{'$anchor'})
-    if not $self->_is_type('string', $schema->{'$anchor'});
+  assert_keyword_type($state, $schema, 'string');
 
   if ($schema->{'$anchor'} !~ /^[A-Za-z][A-Za-z0-9_:.-]+$/) {
     $self->_remove_resource($state->{base_uri}->clone->fragment($schema->{'$anchor'}));
-    abort($state, '%s does not match required syntax', $schema->{'$anchor'});
+    abort($state, '$anchor value "%s" does not match required syntax', $schema->{'$anchor'});
   }
 
   # we already indexed this uri, so there is nothing more to do.
-  # we explicitly do NOT set $state->{absolute_schema_uri}.
+  # we explicitly do NOT set $state->{canonical_schema_uri}.
   return 1;
 }
 
 sub _eval_keyword_ref {
   my ($self, $data, $schema, $state) = @_;
 
+  assert_keyword_type($state, $schema, 'string');
+
   my $uri = Mojo::URL->new($schema->{'$ref'})->base($state->{base_uri})->to_abs;
 
   my $fragment = $uri->fragment // '';
-  my ($subschema, $absolute_uri);
+  my ($subschema, $canonical_uri);
   # TODO: this will get less ugly when we move to actual document objects
   if (not length($fragment) or $fragment =~ m{^/}) {
     my $base = $uri->clone->fragment(undef);
     my $document = Mojo::JSON::Pointer->new(($self->_get_resource($base) // {})->{ref});
     $subschema = $document->get($fragment);
-    $absolute_uri = $uri;
+    $canonical_uri = $uri;
   }
   else {
     if (my $resource = $self->_get_resource($uri)) {
       $subschema = $resource->{ref};
-      $absolute_uri = $resource->{canonical_uri}->clone;  # this is *not* the anchor-containing URI
+      $canonical_uri = $resource->{canonical_uri}->clone; # this is *not* the anchor-containing URI
     }
   }
 
@@ -261,11 +265,41 @@ sub _eval_keyword_ref {
   return $self->_eval($data, $subschema,
     +{ %$state,
       traversed_schema_path => $state->{traversed_schema_path}.$state->{schema_path}.'/$ref',
-      absolute_schema_uri => $absolute_uri,
+      canonical_schema_uri => $canonical_uri, # note: not canonical yet until $id is processed
       schema_path => '',
     });
 }
 
+sub _eval_keyword_vocabulary {
+  my ($self, $data, $schema, $state) = @_;
+
+  assert_keyword_type($state, $schema, 'object');
+
+  # we do nothing with this keyword yet. When we know we are in a metaschema,
+  # we can scan the URIs included here and either abort if a vocabulary is enabled that we do not
+  # understand, or turn on and off certain keyword behaviours based on the boolean values seen.
+
+  return 1;
+}
+
+sub _eval_keyword_comment {
+  my ($self, $data, $schema, $state) = @_;
+  assert_keyword_type($state, $schema, 'string');
+  # we do nothing with this keyword, including not collecting its value for annotations.
+  return 1;
+}
+
+sub _eval_keyword_defs {
+  my ($self, $data, $schema, $state) = @_;
+
+  my $type = $self->_get_type($schema->{'$defs'});
+  abort($state, '$defs value is not an object or boolean')
+    if $type ne 'object' and $type ne 'boolean';
+
+  # we do nothing directly with this keyword, including not collecting its value for annotations.
+  return 1;
+}
+
 sub _eval_keyword_type {
   my ($self, $data, $schema, $state) = @_;
 
@@ -302,9 +336,8 @@ sub _eval_keyword_multipleOf {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('number', $data);
-  abort($state, '%s is not a number', $schema->{multipleOf})
-    if not $self->_is_type('number', $schema->{multipleOf});
-  abort($state, '%s is not a positive number', $schema->{multipleOf}) if $schema->{multipleOf} <= 0;
+  assert_keyword_type($state, $schema, 'number');
+  abort($state, 'multipleOf value is not a positive number') if $schema->{multipleOf} <= 0;
 
   my $quotient = $data / $schema->{multipleOf};
   return 1 if int($quotient) == $quotient;
@@ -315,8 +348,7 @@ sub _eval_keyword_maximum {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('number', $data);
-  abort($state, '%s is not a number', $schema->{maximum})
-    if not $self->_is_type('number', $schema->{maximum});
+  assert_keyword_type($state, $schema, 'number');
 
   return 1 if $data <= $schema->{maximum};
   return E($state, 'value is larger than %d', $schema->{maximum});
@@ -326,8 +358,7 @@ sub _eval_keyword_exclusiveMaximum {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('number', $data);
-  abort($state, '%s is not a number', $schema->{exclusiveMaximum})
-    if not $self->_is_type('number', $schema->{exclusiveMaximum});
+  assert_keyword_type($state, $schema, 'number');
 
   return 1 if $data < $schema->{exclusiveMaximum};
   return E($state, 'value is equal to or larger than %d', $schema->{exclusiveMaximum});
@@ -337,8 +368,7 @@ sub _eval_keyword_minimum {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('number', $data);
-  abort($state, '%s is not a number', $schema->{minimum})
-    if not $self->_is_type('number', $schema->{minimum});
+  assert_keyword_type($state, $schema, 'number');
 
   return 1 if $data >= $schema->{minimum};
   return E($state, 'value is smaller than %d', $schema->{minimum});
@@ -348,8 +378,7 @@ sub _eval_keyword_exclusiveMinimum {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('number', $data);
-  abort($state, '%s is not a number', $schema->{exclusiveMinimum})
-    if not $self->_is_type('number', $schema->{exclusiveMinimum});
+  assert_keyword_type($state, $schema, 'number');
 
   return 1 if $data > $schema->{exclusiveMinimum};
   return E($state, 'value is equal to or smaller than %d', $schema->{exclusiveMinimum});
@@ -359,10 +388,8 @@ sub _eval_keyword_maxLength {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('string', $data);
-  abort($state, '%s is not an integer', $schema->{maxLength})
-    if not $self->_is_type('integer', $schema->{maxLength});
-  abort($state, '%s is not a non-negative integer', $schema->{maxLength})
-    if $schema->{maxLength} < 0;
+  assert_keyword_type($state, $schema, 'integer');
+  abort($state, 'maxLength value is not a non-negative integer') if $schema->{maxLength} < 0;
 
   return 1 if length($data) <= $schema->{maxLength};
   return E($state, 'length is greater than %d', $schema->{maxLength});
@@ -372,10 +399,8 @@ sub _eval_keyword_minLength {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('string', $data);
-  abort($state, '%s is not an integer', $schema->{minLength})
-    if not $self->_is_type('integer', $schema->{minLength});
-  abort($state, '%s is not a non-negative integer', $schema->{minLength})
-    if $schema->{minLength} < 0;
+  assert_keyword_type($state, $schema, 'integer');
+  abort($state, 'minLength value is not a non-negative integer') if $schema->{minLength} < 0;
 
   return 1 if length($data) >= $schema->{minLength};
   return E($state, 'length is less than %d', $schema->{minLength});
@@ -385,6 +410,7 @@ sub _eval_keyword_pattern {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('string', $data);
+  assert_keyword_type($state, $schema, 'string');
 
   return 1 if $data =~ qr/$schema->{pattern}/;
   return E($state, 'pattern does not match');
@@ -394,10 +420,8 @@ sub _eval_keyword_maxItems {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('array', $data);
-  abort($state, '%s is not an integer', $schema->{maxItems})
-    if not $self->_is_type('integer', $schema->{maxItems});
-  abort($state, '%s is not a non-negative integer', $schema->{maxItems})
-    if $schema->{maxItems} < 0;
+  assert_keyword_type($state, $schema, 'integer');
+  abort($state, 'maxItems value is not a non-negative integer') if $schema->{maxItems} < 0;
 
   return 1 if @$data <= $schema->{maxItems};
   return E($state, 'more than %d items', $schema->{maxItems});
@@ -407,10 +431,8 @@ sub _eval_keyword_minItems {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('array', $data);
-  abort($state, '%s is not an integer', $schema->{minItems})
-    if not $self->_is_type('integer', $schema->{minItems});
-  abort($state, '%s is not a non-negative integer', $schema->{minItems})
-    if $schema->{minItems} < 0;
+  assert_keyword_type($state, $schema, 'integer');
+  abort($state, 'minItems value is not a non-negative integer') if $schema->{minItems} < 0;
 
   return 1 if @$data >= $schema->{minItems};
   return E($state, 'fewer than %d items', $schema->{minItems});
@@ -420,8 +442,7 @@ sub _eval_keyword_uniqueItems {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('array', $data);
-  abort($state, '%s is not a boolean', $schema->{uniqueItems})
-    if not $self->_is_type('boolean', $schema->{uniqueItems});
+  assert_keyword_type($state, $schema, 'boolean');
 
   return 1 if not $schema->{uniqueItems};
   return 1 if $self->_is_elements_unique($data, my $equal_indices = []);
@@ -432,9 +453,8 @@ sub _eval_keyword_maxProperties {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('object', $data);
-  abort($state, '%s is not an integer', $schema->{maxProperties})
-    if not $self->_is_type('integer', $schema->{maxProperties});
-  abort($state, '%s is not a non-negative integer', $schema->{maxProperties})
+  assert_keyword_type($state, $schema, 'integer');
+  abort($state, 'maxProperties value is not a non-negative integer')
     if $schema->{maxProperties} < 0;
 
   return 1 if keys %$data <= $schema->{maxProperties};
@@ -445,9 +465,8 @@ sub _eval_keyword_minProperties {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('object', $data);
-  abort($state, '%s is not an integer', $schema->{minProperties})
-    if not $self->_is_type('integer', $schema->{minProperties});
-  abort($state, '%s is not a non-negative integer', $schema->{minProperties})
+  assert_keyword_type($state, $schema, 'integer');
+  abort($state, 'minProperties value is not a non-negative integer')
     if $schema->{minProperties} < 0;
 
   return 1 if keys %$data >= $schema->{minProperties};
@@ -458,8 +477,7 @@ sub _eval_keyword_required {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('object', $data);
-  abort($state, '"required" value is not an array')
-    if not $self->_is_type('array', $schema->{required});
+  assert_keyword_type($state, $schema, 'array');
   abort($state, '"required" element is not a string')
     if any { !$self->_is_type('string', $_) } @{$schema->{required}};
 
@@ -472,8 +490,7 @@ sub _eval_keyword_dependentRequired {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('object', $data);
-  abort($state, '"dependentRequired" value is not an object')
-    if not $self->_is_type('object', $schema->{dependentRequired});
+  assert_keyword_type($state, $schema, 'object');
   abort($state, '"dependentRequired" property is not an array')
     if any { !$self->_is_type('array', $schema->{dependentRequired}{$_}) }
       keys %{$schema->{dependentRequired}};
@@ -486,13 +503,13 @@ sub _eval_keyword_dependentRequired {
     keys %{$schema->{dependentRequired}};
 
   return 1 if not @missing;
-  return E($state, 'missing propert'.(@missing > 1 ? 'ies' : 'y').': '.join(', ', @missing));
+  return E($state, 'missing propert'.(@missing > 1 ? 'ies' : 'y').': '.join(', ', sort @missing));
 }
 
 sub _eval_keyword_allOf {
   my ($self, $data, $schema, $state) = @_;
 
-  abort($state, '"allOf" value is not an array') if not $self->_is_type('array', $schema->{allOf});
+  assert_keyword_type($state, $schema, 'array');
   abort($state, '"allOf" array is empty') if not @{$schema->{allOf}};
 
   my @invalid;
@@ -512,7 +529,7 @@ sub _eval_keyword_allOf {
 sub _eval_keyword_anyOf {
   my ($self, $data, $schema, $state) = @_;
 
-  abort($state, '"anyOf" value is not an array') if not $self->_is_type('array', $schema->{anyOf});
+  assert_keyword_type($state, $schema, 'array');
   abort($state, '"anyOf" array is empty') if not @{$schema->{anyOf}};
 
   my $valid = 0;
@@ -532,7 +549,7 @@ sub _eval_keyword_anyOf {
 sub _eval_keyword_oneOf {
   my ($self, $data, $schema, $state) = @_;
 
-  abort($state, '"oneOf" value is not an array') if not $self->_is_type('array', $schema->{oneOf});
+  assert_keyword_type($state, $schema, 'array');
   abort($state, '"oneOf" array is empty') if not @{$schema->{oneOf}};
 
   my (@valid, @errors);
@@ -588,11 +605,10 @@ sub _eval_keyword_dependentSchemas {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('object', $data);
-  abort($state, '"dependentSchemas" value is not an object')
-    if not $self->_is_type('object', $schema->{dependentSchemas});
+  assert_keyword_type($state, $schema, 'object');
 
   my $valid = 1;
-  foreach my $property (keys %{$schema->{dependentSchemas}}) {
+  foreach my $property (sort keys %{$schema->{dependentSchemas}}) {
     next if not exists $data->{$property}
       or $self->_eval($data, $schema->{dependentSchemas}{$property},
         +{ %$state, schema_path => $state->{schema_path}.'/dependentSchemas/'.$property });
@@ -691,10 +707,9 @@ sub _eval_keyword_contains {
   }
 
   if (exists $schema->{minContains}) {
-    abort($state, '%s is not an integer', $schema->{minContains})
-      if not $self->_is_type('integer', $schema->{minContains});
-    abort($state, '%s is not a non-negative integer', $schema->{minContains})
-      if $schema->{minContains} < 0;
+    local $state->{keyword} = 'minContains';
+    assert_keyword_type($state, $schema, 'integer');
+    abort($state, 'minContains value is not a non-negative integer') if $schema->{minContains} < 0;
   }
 
   my $valid = 1;
@@ -707,14 +722,13 @@ sub _eval_keyword_contains {
   }
 
   if (exists $schema->{maxContains}) {
-    abort($state, '%s is not an integer', $schema->{maxContains})
-      if not $self->_is_type('integer', $schema->{maxContains});
-    abort($state, '%s is not a non-negative integer', $schema->{maxContains})
-      if $schema->{maxContains} < 0;
+    local $state->{keyword} = 'maxContains';
+    assert_keyword_type($state, $schema, 'integer');
+    abort($state, 'maxContains value is not a non-negative integer') if $schema->{maxContains} < 0;
 
     if ($num_valid > $schema->{maxContains}) {
       $valid = 0;
-      E({ %$state, keyword => 'maxContains' }, 'contains too many matching items');
+      E($state, 'contains too many matching items');
       return 0 if $state->{short_circuit};
     }
   }
@@ -732,11 +746,10 @@ sub _eval_keyword_properties {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('object', $data);
-  abort($state, '"properties" value is not an object')
-    if not $self->_is_type('object', $schema->{properties});
+  assert_keyword_type($state, $schema, 'object');
 
   my $valid = 1;
-  foreach my $property (keys %{$schema->{properties}}) {
+  foreach my $property (sort keys %{$schema->{properties}}) {
     next if not exists $data->{$property};
     $valid = 0 if not $self->_eval($data->{$property}, $schema->{properties}{$property},
         +{ %$state,
@@ -754,13 +767,11 @@ sub _eval_keyword_patternProperties {
   my ($self, $data, $schema, $state) = @_;
 
   return 1 if not $self->_is_type('object', $data);
-  abort($state, '"patternProperties" value is not an object')
-    if not $self->_is_type('object', $schema->{patternProperties});
+  assert_keyword_type($state, $schema, 'object');
 
   my $valid = 1;
-  foreach my $property_pattern (keys %{$schema->{patternProperties}}) {
-    my @property_matches = grep /$property_pattern/, keys %$data;
-    foreach my $property (@property_matches) {
+  foreach my $property_pattern (sort keys %{$schema->{patternProperties}}) {
+    foreach my $property (sort grep /$property_pattern/, keys %$data) {
       $valid = 0
         if not $self->_eval($data->{$property}, $schema->{patternProperties}{$property_pattern},
           +{ %$state,
@@ -781,7 +792,7 @@ sub _eval_keyword_additionalProperties {
   return 1 if not $self->_is_type('object', $data);
 
   my $valid = 1;
-  foreach my $property (keys %$data) {
+  foreach my $property (sort keys %$data) {
     next if exists $schema->{properties} and exists $schema->{properties}{$property};
     next if exists $schema->{patternProperties}
       and any { $property =~ /$_/ } keys %{$schema->{patternProperties}};
@@ -816,7 +827,7 @@ sub _eval_keyword_propertyNames {
   return 1 if not $self->_is_type('object', $data);
 
   my $valid = 1;
-  foreach my $property (keys %$data) {
+  foreach my $property (sort keys %$data) {
     $valid = 0 if not $self->_eval($property, $schema->{propertyNames},
       +{ %$state,
         data_path => $state->{data_path}.'/'.$property,
@@ -830,7 +841,7 @@ sub _eval_keyword_propertyNames {
 }
 
 sub _is_type {
-  my ($self, $type, $value) = @_;
+  my (undef, $type, $value) = @_;
 
   if ($type eq 'null') {
     return !(defined $value);
@@ -948,8 +959,10 @@ sub _traverse_for_identifiers {
     if (exists $data->{'$id'} and _is_type(undef, 'string', $data->{'$id'})) {
       $canonical_uri = Mojo::URL->new($data->{'$id'})->base($canonical_uri)->to_abs;
       # this might not be a real $id... wait for it to be encountered at runtime before dying
-      $identifiers{$canonical_uri} = { ref => $data, canonical_uri => $canonical_uri }
-        if not length $canonical_uri->fragment;
+      if (not length $canonical_uri->fragment) {
+        $canonical_uri->fragment(undef);
+        $identifiers{$canonical_uri} = { ref => $data, canonical_uri => $canonical_uri };
+      };
     }
     if (exists $data->{'$anchor'} and _is_type(undef, 'string', $data->{'$anchor'})) {
       # we cannot change the canonical uri, or we won't be able to properly identify
@@ -989,9 +1002,11 @@ sub E {
   push @{$state->{errors}}, JSON::Schema::Draft201909::Error->new(
     instance_location => $state->{data_path},
     keyword_location => $state->{traversed_schema_path}.$state->{schema_path}.$suffix,
-    !$state->{absolute_schema_uri} ? () : ( absolute_keyword_location => do {
-      my $abs = $state->{absolute_schema_uri}->clone;
-      $abs->fragment(($abs->fragment//'').$state->{schema_path}.$suffix);
+    !$state->{canonical_schema_uri} ? () : ( absolute_keyword_location => do {
+      my $uri = $state->{canonical_schema_uri}->clone;
+      $uri->fragment(($uri->fragment//'').$state->{schema_path}.$suffix)
+        if $state->{schema_path} or $suffix;
+      $uri;
     } ),
     error => @args ? sprintf($error_string, @args) : $error_string,
   );
@@ -1007,6 +1022,14 @@ sub abort {
   die pop @{$state->{errors}};
 }
 
+# one common usecase of abort()
+use namespace::clean 'assert_keyword_type';
+sub assert_keyword_type {
+  my ($state, $schema, $type) = @_;
+  abort($state, $state->{keyword}.' value is not a%s %s', ($type =~ /^[aeiou]/ ? 'n' : ''), $type)
+    if not _is_type(undef, $type, $schema->{$state->{keyword}});
+}
+
 1;
 
 __END__
@@ -1023,7 +1046,7 @@ JSON::Schema::Draft201909 - Validate data against a schema
 
 =head1 VERSION
 
-version 0.002
+version 0.003
 
 =head1 SYNOPSIS
 
@@ -1053,6 +1076,12 @@ than continuing to find all errors.
 
 Defaults to true when C<output_format> is C<flag>, and false otherwise.
 
+=head2 max_traversal_depth
+
+The maximum number of levels deep a schema traversal may go, before evaluation is halted. This is to
+protect against accidental infinite recursion, such as from two subschemas that each reference each
+other. Defaults to 50.
+
 =head1 METHODS
 
 =head2 evaluate_json_string
@@ -1093,6 +1122,8 @@ used in another context (for example, a numeric value is concatenated into a str
 string is used in an arithmetic operation), additional flags can be added onto the variable causing
 it to resemble the other type. This should not be an issue if data validation is occurring
 immediately after decoding a JSON payload, or if the JSON string itself is passed to this module.
+If this turns out to be an issue in real environments, I may have to implement a C<lax_scalars>
+option.
 
 For more information, see L<Cpanel::JSON::XS/MAPPING>.
 
