@@ -4,7 +4,7 @@ package JSON::Schema::Draft201909::Error;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Contains a single error from a JSON Schema evaluation
 
-our $VERSION = '0.002';
+our $VERSION = '0.003';
 
 no if "$]" >= 5.031009, feature => 'indirect';
 use Moo;
@@ -47,7 +47,7 @@ __END__
 
 =encoding UTF-8
 
-=for stopwords schema
+=for stopwords schema fragmentless
 
 =head1 NAME
 
@@ -55,7 +55,7 @@ JSON::Schema::Draft201909::Error - Contains a single error from a JSON Schema ev
 
 =head1 VERSION
 
-version 0.002
+version 0.003
 
 =head1 SYNOPSIS
 
@@ -87,10 +87,10 @@ The schema path taken during evaluation to arrive at the error.
 =head2 absolute_keyword_location
 
 The path in the schema where the error occurred (may be different from keyword_location, if a
-C<$ref> was followed).  This is supposed to be an absolute URI (as per
-L<https://json-schema.org/draft/2019-09/json-schema-core.html#rfc.section.10.3.2>) but
-this implementation does not yet track the absolute URIs of schemas, so it is just the fragment
-portion of a URI for now.
+C<$ref> was followed).  Note that this is not actually an absolute (fragmentless) URI in most cases,
+as the indicated error will occur at a path below the position where the most recent identifier had
+been declared in the schema. Further, if the schema never declared an absolute base URI, this URI
+won't contain a scheme either.
 
 =head2 error
 
